@@ -1,5 +1,5 @@
 ;; scss.scm: main module exports and implementations for SCSS
-;; Copyright (C) 2005 Julian Graham
+;; Copyright (C) 2006 Julian Graham
 
 ;; SCSS is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -46,7 +46,8 @@
 	    scss:get-default-value
 
 	    scss:select-value
-	    scss:select-value-at-node))
+	    scss:select-value-at-node
+	    scss:clear-style-cache!))
 
 (load "lexer.scm")
 (load "parser.scm")
@@ -514,6 +515,7 @@
 					    (get-sxml-parent x node))
 					  c))
 			      #f))))
+
 (define sxml-node-name
   (lambda (node)
     (let* ((str (false-if-exception (symbol->string (car node))))
@@ -625,8 +627,10 @@
 		     (set! read-ident #t))
 		    ((and read-ident (or (equal? "+" y) (equal? ">" y)))
 		     (set! read-ident #f))
-		    ((or (and read-ident (or (pseudo-class? y) 
-					     (pseudo-element? y)))
+		    ((or (and read-ident 
+			      (or (pseudo-class? y) 
+				  (and (pseudo-element? y)
+				       (equal? (car (last-pair x)) y))))
 			 (exact-exec id-regexc y)
 			 (exact-exec dot-regexc y)))
 		    (else (error #t 
@@ -762,7 +766,13 @@
 					       (string-append bu "/" s1))))))
 			      (new-sheet (false-if-exception (scss:css->scss 
 							      p nbu))))
-			 (if new-sheet (append new-sheet s2) s2)))
+			 (if new-sheet 
+			     (append (map (lambda (x)
+					    (cons (list (car x))
+						  (cdr x)))
+					  new-sheet)
+				     s2) 
+			     (list s2))))
 		     clean-sheet
 		     imports)))
 	(expand-stylesheet final)))))
@@ -821,7 +831,6 @@
 		   arglist)
 	    (error #t "wrong number of arguments to scss:create-cascade")))))
     
-
 (define scss:create-ruleset
   (lambda (sel-str . rest)
     (verify-arg-types "scss:create-ruleset"
@@ -842,7 +851,6 @@
     (let ((sel (parse-selector sel-str)))
       (list sel rest))))
 						     
-
 (define scss:set-author-stylesheet!
   (lambda (cascade authorsheet) 
     (set-car! (list-cdr-ref cascade 2) authorsheet)))
@@ -946,13 +954,31 @@
        stylesheet)
       ret)))
 
+(define get-more-general-properties
+  (lambda (prop)
+    (let ((prop-bits (string-split prop #\-)))
+      (cond ((<= (length prop-bits) 1) (list))
+	    ((member (car prop-bits) 
+		     '("background" "margin" "outline" "padding"))
+	     (list (car prop-bits)))
+	    ((and (equal? (car prop-bits) "list")
+		  (equal? (cadr prop-bits) "style"))
+	     (list "list-style"))
+	    ((equal? (car prop-bits) "border")
+	     (if (eqv? (length prop-bits) 2) 
+		 (list "border")
+		 (list (string-append "border-" (cadr prop-bits))
+		       (string-append "border-" (caddr prop-bits))
+		       "border")))
+	    (else (list))))))
+
 ;; This function is a bit complicated.  The general idea is that for certain
 ;; properties, if they're not explicitly defined in the matched selector 
 ;; block, you can derive their value by looking at a more general property 
 ;; that is defined in that block.
 
 (define get-general-property
-  (lambda (prop ph oa)
+  (lambda (sprop val oa)
     (let* ((split-matches 
 	    (lambda (pattern str)
 	      (let* ((lst (list))
@@ -962,7 +988,7 @@
 		(regexp-substitute/global #f pattern str f 'post)
 		lst)))
 
-	   (prop-bits (string-split prop #\-))
+	   (prop-bits (string-split sprop #\-))
 
 	   ;; The way values are mapped to specific properties depends on the 
 	   ;; number of values as per 
@@ -991,120 +1017,63 @@
       (cond ((= (length prop-bits) 1) #f)
 	    ((equal? (car prop-bits) "border")
 	     (if (= (length prop-bits) 2)
-		 (let ((b (hash-ref ph "border")))
-		   (if b 
-		       (cond ((equal? (cadr prop-bits) "color") 
-			      (let ((res (h (car b) color-regex)))
-				(if res res (cl))))
-			     ((equal? (cadr prop-bits) "width")
-			      (h (car b) border-width-regex))
-			     ((equal? (cadr prop-bits) "style")
-			      (h (car b) border-style-regex))
-
-			     ;; Else it's one of top, left, bottom, right
-
-			     (else (car b)))
-		       #f))
+		 (cond ((equal? (cadr prop-bits) "color") 
+			(let ((res (h val color-regex)))
+			  (if res res (cl))))
+		       ((equal? (cadr prop-bits) "width")
+			(h val border-width-regex))
+		       ((equal? (cadr prop-bits) "style")
+			(h val border-style-regex))
+		       
+		       ;; Else it's one of top, left, bottom, right
+		       
+		       (else (car b)))
 		 (let* ((c (caddr prop-bits))
 			(d (string-append (car prop-bits) 
-					  "-" (cadr prop-bits)))
-			(a (hash-ref ph (string-append "border-" c)))
-			(b (hash-ref ph d))
-			(a (if (and a b (> (cdr a) (cdr b)))
-			       #f (if a (car a) a)))
-			(b (if (and a b (number? a) (> (cdr b) (cdr a))) 
-			       #f (if b (car b) b))))
-		   (cond (a (cond ((equal? (caddr prop-bits) "color")
-				   (let ((res (g color-regex (cadr prop-bits) 
-						 a)))
-				     (if res res (cl)))) 
-				  ((equal? (caddr prop-bits) "width")
-				   (g border-width-regex (cadr prop-bits) a))
-				  ((equal? (caddr prop-bits) "style")
-				   (g border-style-regex (cadr prop-bits) 
-				      a))))
-			 (b (cond ((equal? (caddr prop-bits) "color")
-				   (let ((res (h b color-regex)))
-				     (if res res (cl))))
-				  ((equal? (caddr prop-bits) "width")
-				   (h b border-width-regex))
-				  ((equal? (caddr prop-bits) "style")
-				   (h b border-style-regex))))
-			 (else (let ((e (get-general-property d ph oa)))
-				 (if e
-				     (cond ((equal? c "color") 
-					    (let ((res (h e color-regex)))
-					      (if res res (cl))))
-					   ((equal? c "width") 
-					    (h e border-width-regex))
-					   ((equal? c "style")
-					    (h e border-style-regex)))
-				     (if (equal? c "color") (cl) #f))))))))
+					  "-" (cadr prop-bits))))
+		   (cond ((equal? (caddr prop-bits) "color")
+			  (let ((res (g color-regex (cadr prop-bits) val)))
+			    (if res res (cl))))
+			 ((equal? (caddr prop-bits) "width")
+			  (g border-width-regex (cadr prop-bits) val))
+			 ((equal? (caddr prop-bits) "style")
+			  (g border-style-regex (cadr prop-bits) val))))))
 	    ((equal? (car prop-bits) "padding")
-	     (let ((p (hash-ref ph "padding"))) 
-	       (if p (g padding-regex (cadr prop-bits) (car p)) #f)))
+	     (g padding-regex (cadr prop-bits) val))
 	    ((equal? (car prop-bits) "margin")
-	     (let ((m (hash-ref ph "margin"))) 
-	       (if m (g margin-regex (cadr prop-bits) (car m)) #f)))
+	     (g margin-regex (cadr prop-bits) val))
 	    ((equal? (car prop-bits) "background")
-	     (let ((b (hash-ref ph "background")))
-	       (if b 
-		   (cond ((equal? (cadr prop-bits) "color") 
-			  (h (car b) pcolor-regex))
-			 ((equal? (cadr prop-bits) "attach") 
-			  (h (car b) pattach-regex))
-			 ((equal? (cadr prop-bits) "image") 
-			  (h (car b) pimage-regex))
-			 ((equal? (cadr prop-bits) "repeat") 
-			  (h (car b) prepeat-regex))
-			 ;; FIX THIS!!!
-			 ((equal? (cadr prop-bits) "position"))
-			 (else #f))
-		   #f)))
+	     (cond ((equal? (cadr prop-bits) "color") (h val pcolor-regex))
+		   ((equal? (cadr prop-bits) "attach") (h val pattach-regex))
+		   ((equal? (cadr prop-bits) "image") (h val pimage-regex))
+		   ((equal? (cadr prop-bits) "repeat") (h val prepeat-regex))
+		   ;; FIX THIS!!!
+		   ((equal? (cadr prop-bits) "position"))
+		   (else #f)))
 	    ((and (equal? (car prop-bits) "list")
 		  (equal? (cadr prop-bits) "style"))
-	     (let ((l (hash-ref ph "list-style")))
-	       (if l
-		   (cond ((equal? (caddr prop-bits) "position")
-			  (h (car l) 
-			     (string-append lsposition-regex "|(inherit)")))
-			 ((equal? (caddr prop-bits) "image")
-			  (h (car l)
-			     (string-append lsimage-regex "|(inherit)")))
-			 ((equal? (caddr prop-bits) "type")
-			  (h (car l) 
-			     (string-append lstype-regex "|(inherit)")))
-			 (else #f))
-		   #f)))
+	     (cond ((equal? (caddr prop-bits) "position")
+		    (h val (string-append lsposition-regex "|(inherit)")))
+		   ((equal? (caddr prop-bits) "image")
+		    (h val (string-append lsimage-regex "|(inherit)")))
+		   ((equal? (caddr prop-bits) "type")
+		    (h val (string-append lstype-regex "|(inherit)")))
+		   (else #f)))
 	    ((equal? (car prop-bits) "outline")
-	     (let ((o (hash-ref ph "outline")))
-	       (if o
-		   (cond ((equal? (cadr prop-bits) "color")
-			  (h (car o) 
-			     (string-append ocolor-regex "|(inherit)")))
-			 ((equal? (cadr prop-bits) "style")
-			  (h (car o) 
-			     (string-append ostyle-regex "|(inherit)")))
-			 ((equal? (cadr prop-bits) "width")
-			  (h (car o)
-			     (string-append owidth-regex "|(inherit)")))
-			 (else #f))
-		   #f)))
+	     (cond ((equal? (cadr prop-bits) "color")
+		    (h val (string-append ocolor-regex "|(inherit)")))
+		   ((equal? (cadr prop-bits) "style")
+		    (h val (string-append ostyle-regex "|(inherit)")))
+		   ((equal? (cadr prop-bits) "width")
+		    (h val (string-append owidth-regex "|(inherit)")))
+		   (else #f)))
 	    ((equal? (car prop-bits) "font")
-	     (let ((font (hash-ref ph "font")))
-	       (if font
-		   (cond ((equal? (cadr prop-bits) "family") 
-			  (h (car font) ffamily-regex))
-			 ((equal? (cadr prop-bits) "size") 
-			  (h (car font) fsize-regex))
-			 ((equal? (cadr prop-bits) "style") 
-			  (h (car font) fstyle-regex))
-			 ((equal? (cadr prop-bits) "variant")
-			  (h (car font) fvariant-regex))
-			 ((equal? (cadr prop-bits) "weight") 
-			  (h (car font) fweight-regex))
-			 (else #f))
-		   #f)))
+	     (cond ((equal? (cadr prop-bits) "family") (h val ffamily-regex))
+		   ((equal? (cadr prop-bits) "size") (h val fsize-regex))
+		   ((equal? (cadr prop-bits) "style") (h val fstyle-regex))
+		   ((equal? (cadr prop-bits) "variant") (h val fvariant-regex))
+		   ((equal? (cadr prop-bits) "weight") (h val fweight-regex))
+		   (else #f)))
 	    (else #f)))))
 
 (define lookup-values
@@ -1124,20 +1093,18 @@
 	   (user-matches (i 'user (cadr cascade)))
 	   (author-matches (i 'author (caddr cascade)))
 	   (results (append '() agent-matches user-matches author-matches))
-	   (sr (sort (sort results (lambda (x y) (< (calc-precedence x)
-						    (calc-precedence y))))
-		     (lambda (x y) (< (calc-specificity (cadr x))
-				      (calc-specificity (cadr y))))))
+	   (sr (sort results (lambda (x y) 
+			       (and (<= (calc-precedence x)
+					(calc-precedence y))
+				    (<= (calc-specificity (cadr x))
+					(calc-specificity (cadr y)))))))
 	   (order 0))
       (for-each (lambda (x) 
-;		  (display "setting ") (display (list-ref x 2))
-;		  (display " ") (display (list-ref x 3)) (newline)
-		  (let ((sdcy (string-downcase (caddr x))))
+		  (let ((sdcy (string-downcase (caddr x)))
+			(sdcv (string-downcase (cadddr x))))		    
 		    (if (hash-ref property-hash-table sdcy)
 			(begin
-			  (hash-set! table 
-				     sdcy 
-				     (cons (string-downcase (cadddr x)) order))
+			  (hash-set! table sdcy (cons sdcy (cons sdcv order)))
 			  (set! order (+ order 1))))))
 		sr))))
 
@@ -1230,7 +1197,7 @@
 				(not t)))
 		     (selector-match? (cdr sel) d nodes)
 		     #f)))
-	      ((equal? csel snn)
+	      ((or (eqv? srcsel #\*) (equal? csel snn))
 	       (let ((i (list-index 
 			 (lambda (x)
 			   (let ((nix (next-ident (cdr sel))))
@@ -1242,7 +1209,6 @@
 		       (i (selector-match? (cdr sel) d 
 					   (list-cdr-ref (cdr nodes) i)))
 		       (else #f))))
-	      ((eqv? srcsel #\*) (selector-match? (cdr sel) d (cdr nodes)))
 	      ((null? sel) #t)
 	      (else #f))))))
 
@@ -1262,24 +1228,25 @@
 	   (agent-matches (i 'agent (car cascade)))
 	   (user-matches (i 'user (cadr cascade)))
 	   (author-matches (i 'author (caddr cascade)))
-	   (results (append '() agent-matches user-matches author-matches))
-	   (sr (sort (sort results (lambda (x y) (< (calc-precedence x)
-						    (calc-precedence y))))
-		     (lambda (x y) (< (calc-specificity (cadr x)) 
-				      (calc-specificity (cadr y))))))
+	   (results (append '() author-matches user-matches agent-matches))
+	   (sr (sort results 
+		     (lambda (x y)
+		       (let ((cpx (calc-precedence x))
+			     (cpy (calc-precedence y)))
+			 (or (< cpx cpy)
+			     (and (eqv? cpx cpy)
+				  (<= (calc-specificity (cadr x))
+				      (calc-specificity (cadr y)))))))))
 	   (order 0))
       (for-each (lambda (x) 
-;		  (display "setting ") (display (list-ref x 2))
-;		  (display " ") (display (list-ref x 3)) (newline)
-		  (let ((sdcy (string-downcase (caddr x))))
+		  (let ((sdcy (string-downcase (caddr x)))
+			(sdcv (string-downcase (cadddr x))))
 		    (if (hash-ref property-hash-table sdcy)
 			(begin
-			  (hash-set! table 
-				     sdcy 
-				     (cons (string-downcase (cadddr x)) order))
+			  (hash-set! table sdcy (cons sdcy (cons sdcv order)))
 			  (set! order (+ order 1))))))
 		sr))))
-
+  
 (define get-default-prop-value
   (lambda (pe) 
     (if (or (not pe) (null? pe))
@@ -1322,15 +1289,21 @@
 			(lookup-values source selector eh)
 			(hash-set! d selector eh)
 			eh)))
-	   (v (hash-ref e prop))
-	   (v (if v 
-		  (car v) 
-		  (get-general-property prop e (list source selector)))))
+	   (vs (sort (cons (hash-ref e prop)
+			   (map (lambda (z) (hash-ref e z))
+				(get-more-general-properties prop)))
+		     (lambda (x y) 
+		       (or (not y) (and x (> (cddr x) (cddr y)))))))
+	   (v (if (car vs)
+		  (if (equal? (caar vs) prop)
+		      (cadar vs)
+		      (get-general-property prop 
+					    (cadar vs) 
+					    (list source selector)))
+		  #f)))
       (if v 
 	  (string-downcase v) 
-	  (if recurse 
-	      (get-default-prop-value (get-prop-entry prop)) 
-	      '())))))
+	  (if recurse (get-default-prop-value (get-prop-entry prop)) '())))))
 
 (define select-value
   (lambda (source selector recurse doc prop)
@@ -1349,21 +1322,37 @@
 			    (hashq-set! c elt eh)
 			    eh)))
 	   (pe (get-prop-entry (string-downcase prop))))
-      (letrec ((f (lambda (ht elt)
-		    (let* ((e (hashq-ref c elt))
-			   (e (if e e (create-hash elt)))
-			   (v (hash-ref e (string-downcase prop)))
-			   (v (if v 
-				  (car v) 
-				  (get-general-property 
-				   prop e (list source elt doc))))
-			   (par (get-sxml-parent d elt)))
-		      (if (and recurse
-			       (and par (not (null? par)))
-			       (or (and (not v) pe (list-ref pe 2))
-				   (equal? v "inherit")))
-			  (f ht par)
-			  v)))))
+      (letrec 
+	  ((f (lambda (ht elt)
+		(let* ((e (hashq-ref c elt))
+		       (e (if e e (create-hash elt)))
+		       (sdp (string-downcase prop))
+		       
+		       ;; The most specific property might not be the best
+		       ;; match -- look at ALL the related properties and
+		       ;; derive the requested value if it's not the most
+		       ;; specific.
+		       
+		       (vs (sort (cons (hash-ref e sdp)
+				       (map (lambda (z) (hash-ref e z))
+					    (get-more-general-properties sdp)))
+				 (lambda (x y)
+				   (or (not y)
+				       (and x (> (cddr x) (cddr y)))))))
+		       (v (if (car vs)
+			      (if (equal? (caar vs) prop)
+				  (cadar vs)
+				  (get-general-property prop 
+							(cadar vs)
+							(list source elt doc)))
+			      #f))
+		       (par (get-sxml-parent d elt)))
+		  (if (and recurse
+			   (and par (not (null? par)))
+			   (or (and (not v) pe (list-ref pe 2))
+			       (equal? v "inherit")))
+		      (f ht par)
+		      v)))))
 	(let ((val (f c selector)))
 	  (if val
 	      (string-downcase val)
@@ -1379,4 +1368,14 @@
 (define scss:select-value-at-node
   (lambda (source selector . r)
     (apply (if (string? selector) lookup-value select-value)
-	   (append (list source selector #f) r))))
\ No newline at end of file
+	   (append (list source selector #f) r))))
+
+(define scss:clear-style-cache!
+  (lambda (source selector . r)
+    (if (string? selector)
+	(let ((d (hashx-ref cascade-hash cascade-assoc cascade-hash-table 
+			    source)))
+	  (if d (hash-remove! d selector)))
+	(let* ((d (hashq-ref document-hash-table (car r)))) 
+	  (if d (let ((c (hashx-ref cascade-hash cascade-assoc d source)))
+		  (if c (hashq-remove! c selector)))))))) 
